import org.python.core.PyObject;
import org.python.core.PySlice;
import org.python.util.PythonInterpreter;
import java.util.AbstractCollection;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.SortedSet;

public class JavaSlice {

  // constructors
  public JavaSlice(PySlice sl) {
    PyObject w = swClass().__call__(sl);
    T slice = (T)w.__tojava__(T.class);
    this.slice = slice;
    init(slice.start(), slice.stop(), slice.step());
  }
  public JavaSlice(Integer start, Integer stop, Integer step) {
    this.slice=null;
    init(start, step, stop);
  }
  public JavaSlice(Integer start, Integer stop) {
    this(start, stop, null);
  }
  protected void init(Integer start, Integer stop, Integer step) {
    this.start = start; this.stop = stop; this.step = step;
  }

  private Integer start; public Integer start() { return(start); }
  private Integer step;  public Integer step()  { return(step); }
  private Integer stop;  public Integer stop()  { return(stop); }
  private T slice; public T slice() { return(slice); }
  public int start(int n) {
    if(start == null) return(0);
    else if(start < 0) return(Math.max(n+start, 0));
    else return(Math.min(start, n));
  }
  public int step(int n) {
    if(step == null) return(1);
    else return(step);
  }
  public int stop(int n) {
    if(stop == null) return(n);
    else if(stop < 0) return(Math.max(n+stop, 0));
    else return(Math.min(stop, n));
  }

  public Indices indices(int n) {
    return(new Indices(n));
  }

  public class Indices
      extends AbstractCollection<Integer>
      implements SortedSet<Integer>
  {
    private int n;      public int n()      { return(n); }
    private int startn; public int startn() { return(startn); }
    private int stopn;  public int stopn()  { return(stopn); }
    private int stepn;  public int stepn()  { return(stepn); }
    // Note, we make frequent use of the invariant stepn() != 0
    // In the unlikely event that someone feels inspired to subclass
    //   this class, this property must be maintained (or lots of methods
    //   overrriden).

    Indices(Integer n) {
      super();
      if(step(n) == 0)
        throw new IllegalArgumentException("slice step cannot be 0");
      this.n      = n;
      this.startn = start(n);
      this.stopn  = stop(n);
      this.stepn  = step(n);
    }

    public boolean contains(Object obj) {
      if(!(obj instanceof Number)) return(false);
      int n = ((Number)(obj)).intValue();
      if(stepn() > 0)
        return((startn() <= n) && (n < stopn()) && (((n-startn())%stepn())==0));
      else
        return((startn() >= n) && (n > stopn()) && (((startn()-n)%(-stepn()))==0));
    }
    public boolean containsAll(Collection<?> c) {
      Iterator<?> it = c.iterator();
      while(it.hasNext()) {
        Object v = it.next();
        if(!(v instanceof Number)) return(false);
        if(!contains(((Number)(v)).intValue()))
          return(false);
      }
      return(true);
    }
    public boolean equals(Indices that) {
      int sz = this.size();
      if(sz != that.size()) return(false);
      if(sz == 0) return(true);
      if(sz == 1) return(this.first() == that.first());
      return((this.last() == that.last()) && (this.stepn()  == that.stepn()));
    }
    public boolean isEmpty() { return(stepn()*(stopn()-startn()) <= 0); }
    public Iterator<Integer> iterator() {
      return(new Iterator<Integer>() {
        Integer next = startn();
        public boolean hasNext() {
          return(prec(next, stopn()));
        }
        public Integer next() {
          if(!hasNext())
            throw new NoSuchElementException();
          Integer v = next;
          next = next+stepn();
          return(v);
        }
        public void remove() {
          throw new UnsupportedOperationException();
        }
      });
    }

    public int size() {
      if(isEmpty()) return(0);
      else return(1 + (last() - first())/stepn());
    }

    // methods required for SortedSet
    public Comparator<Integer> comparator() {
      if(stepn() > 0) return(null);
      else return(new Comparator<Integer>() {
        public int compare(Integer x, Integer y) { return(y.compareTo(x)); }
      });
    }

    public Integer first() {
      if((stepn() > 0) && (startn() < stopn())) return(startn());
      else if((stepn() < 0) && (startn() > stopn())) return(startn());
      else throw new NoSuchElementException();
    }

    public SortedSet<Integer> headSet(Integer hiBnd) {
      if((hiBnd == null) || precEq(stopn(), hiBnd)) return(this);
      else return((new JavaSlice(startn(), hiBnd, stepn())).indices(n));
    }

    public Integer last() {
      if(isEmpty()) throw new NoSuchElementException();
      return(align(stopn()-incr()));
    }

    public SortedSet<Integer> subSet(Integer loBnd, Integer hiBnd) {
      return(headSet(hiBnd).tailSet(loBnd));
    }

    public SortedSet<Integer> tailSet(Integer loBnd) {
      if((loBnd == null) || precEq(loBnd, startn())) return(this);
      return((new JavaSlice(align(loBnd), stopn(), stepn())).indices(n));
    }

    int align(int a) {
      if(prec(a, startn()))
        throw new AssertionError("Oops! -- please report this bug");
      if(stepn() > 0) return(a - ((a - startn()) % stepn()));
      else return(a + ((startn()-1) % (-stepn())));
    }
    int incr() { return((stepn() > 0) ? 1 : -1); }
    boolean prec(int a, int b) { // a precedes b in the order of this slice
      return((a-b)*stepn() < 0);
    }
    boolean precEq(int a, int b) {
      return((a == b) || prec(a, b));
    }
  }


  // code for interfacing with Python
  private static PyObject swClass = null;
  protected PyObject swClass() {
    if(swClass == null) {
      String[] SliceWrapper = new String[] {
        "import JavaSlice\n",
        "class SliceWrapper(JavaSlice.T):\n",
        "  def __init__(self, sl): self.sl = sl\n",
        "  def start(self): return(self.sl.start)\n",
        "  def step(self): return(self.sl.step)\n",
        "  def stop(self): return(self.sl.stop)\n"
      };
      PythonInterpreter interpreter = new PythonInterpreter();
      StringBuffer buf = new StringBuffer();
      for(int i = 0; i < SliceWrapper.length; i++)
        buf.append(SliceWrapper[i]);
      interpreter.exec(buf.toString());
      swClass = interpreter.get("SliceWrapper");
    }
    return(swClass);
  }

  public interface T {
    public Integer start();
    public Integer step();
    public Integer stop();
  }
}
