import net.sf.javabdd.*;
import java.util.*;


public class SymFactory {
  // Defaults from the "medium sized example from:
  //    http://buddy.sourceforge.net/manual/main.html
  public final static int    defaultNodeNum    = 160000;
  public final static int    defaultCacheSize  =  10000;
  public final static double defaultCacheRatio =     16.0;

  private BDDFactory bf;
  private Map<String,SymBDD> symTable;
  private boolean done, needToSetVarOrder;
  private Map<Integer,SymBool> varTable;

  public SymFactory(int nodenum, int cachesize) {
    bf = JFactory.init(nodenum, cachesize);
    bf.setCacheRatio(defaultCacheRatio);
    symTable = new HashMap<String,SymBDD>();
    varTable = new HashMap<Integer,SymBool>();
    done = false;
    needToSetVarOrder = false;
    try {
      java.lang.reflect.Method m =
	getClass().getMethod("do_nothing", (java.lang.Class[])null);
      bf.registerGCCallback(this, m);
      bf.registerReorderCallback(this, m);
      bf.registerResizeCallback(this, m);
    } catch (Exception e) {}
  }
  public void do_nothing() {}

  public SymFactory() {
    this(100000, 100000);
  }

  public void done() {
    bf.done();
    done = true;
    if(this == defaultSF)
      defaultSF = null;
  }

  public SymBDD getSym(String name) {
    return(symTable.get(name));
  }

  public void putSym(SymBDD sym) {
    if(symTable.get(sym.name()) != null)
      throw new AlreadyDeclaredException(
        "SymFactory.put: " + sym.name() + " already defined");
    symTable.put(sym.name(), sym);
  }

  public int varNum() {
    return(bf.varNum());
  }

  public int extVarNum(int howMany) {
    return(bf.extVarNum(howMany));
  }

  public void putVarSym(int var, SymBool sb) {
    varTable.put(var, sb);
    needToSetVarOrder = true;
  }

  public SymBool varSym(int var) {
    return(varTable.get(var));
  }
  public String varName(int var) {
    SymBool sb = varSym(var);
    String s = (sb == null) ? null : sb.name();
    if(s == null) {
      if((var < 0) || (var >= bf.varNum()))
        s = "$NoSuchVariable";
      else s = (new Formatter()).format("$%03d", var).toString();
    }
    return(s);
  }

  public void setVarOrder(Comparator<SymBool> c) {
    int[] m = new int[varTable.size()];
    TreeSet<SymBool> ts =
      ( (c == null)
        ? new TreeSet<SymBool>(new Comparator<SymBool>() {
	    public int compare(SymBool x, SymBool y) {
	      int d = x.rank() - y.rank();
	      if(d != 0) return(d);
	      else return(x.vnum() - y.vnum());
	    }
	  })
	: new TreeSet<SymBool>(c)
      );
    ts.addAll(varTable.values());
    Iterator<SymBool> it = ts.iterator();
    for(int i = 0; i < m.length; i++)
      m[i] = it.next().vnum();
    int oldVerbosity = bf.reorderVerbose(0);
    bf.setVarOrder(m);
    bf.reorderVerbose(oldVerbosity);
  }
  public void setVarOrder() { setVarOrder(null); }

  public void setVarOrderIfNeeded() {
    if(needToSetVarOrder) {
      needToSetVarOrder = false;
      setVarOrder();
    }
  }

  public BDD ithVar(int i) {
    return(bf.ithVar(i));
  }

  public BDD nithVar(int i) {
    return(bf.nithVar(i));
  }

  public BDDPairing makePair() { return(bf.makePair()); }

  private SymBool one = null;
  public SymBool one() {
    if(one == null)
      one = new SymBool(bf.one(), this);
    return(one);
  }

  private SymBool zero = null;
  public SymBool zero() {
    if(zero == null)
      zero = new SymBool(bf.zero(), this);
    return(zero);
  }

  public SymBool bool(boolean v) {
    return(v ? one() : zero());
  }
  public SymBool bit(long v, int i) {
    return(bool(((v >> i) & 1) != 0));
  }

  public int level2Var(int level) {
    return(bf.level2Var(level));
  }

  public int nodeCount(Collection<BDD> bdds) {
    return(bf.nodeCount(bdds));
  }

  private static SymFactory defaultSF = null;

  public static SymFactory defaultSF() {
    if(defaultSF == null)
      defaultSF = new SymFactory();
    return(defaultSF);
  }

  public SymBDD do_find(String name) {
    SymBDD sym = getSym(name);
    if(sym != null) return(sym);
    for(int i = 0; i < bf.varNum(); i++) {
      sym = varSym(i);
      if(name.equals(sym.name())) return(sym);
    }
    return(null);
  }

  public static SymBDD find(String name) {
    return(defaultSF().do_find(name));
  }
}
